from social_core.pipeline import social_auth


def social_details(backend, details, response, *args, **kwargs):
    r = social_auth.social_details(backend, details, response, *args, **kwargs)
    r['details']['idp'] = response.get('idp')
    if not r['details'].get('email'):
        if not response.get('email'):
            r['details']['email'] = response["signInNames.emailAddress"]
        else:
            r['details']['email'] = response.get('email')
    email = r['details'].get('email')
    if isinstance(email, str):
        r['details']['email'] = email.lower()
    return r


def get_username(strategy, details, backend, user=None, *args, **kwargs):
    return {'username': details.get('email')}



import os

AZURE_CLIENT_ID = os.environ.get('AZURE_CLIENT_ID', '')
AZURE_CLIENT_SECRET = os.environ.get('AZURE_CLIENT_SECRET', '')

INSIGHT_URL = os.environ.get('INSIGHT_URL', 'https://unibiapitest.azure-api.net/biapi/v1/')
INSIGHT_SUB_KEY = os.environ.get('INSIGHT_SUB_KEY', 'insight_sub_key')

GRAPH_CLIENT_ID = os.environ.get('GRAPH_CLIENT_ID', AZURE_CLIENT_ID)
GRAPH_CLIENT_SECRET = os.environ.get('GRAPH_CLIENT_SECRET', AZURE_CLIENT_SECRET)

AZURE_SSL = True
AZURE_URL_EXPIRATION_SECS = 10800
AZURE_ACCESS_POLICY_EXPIRY = 10800  # length of time before signature expires in seconds
AZURE_ACCESS_POLICY_PERMISSION = 'r'
AZURE_TOKEN_URL = 'https://login.microsoftonline.com/unicef.org/oauth2/token'
AZURE_GRAPH_API_BASE_URL = 'https://graph.microsoft.com'
AZURE_GRAPH_API_VERSION = 'v1.0'
AZURE_GRAPH_API_PAGE_SIZE = 300



import logging

from django.conf import settings
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseServerError
from social_core.exceptions import AuthCanceled, AuthMissingParameter
from social_django.middleware import SocialAuthExceptionMiddleware

from unicef_security.backends import UNICEFAzureADB2COAuth2

logger = logging.getLogger("healthz")


class HealthCheckMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        if request.method == "GET":
            if request.path == "/readiness/":
                return self.readiness(request)
            elif request.path == "/healthz/":
                return self.healthz(request)
        return self.get_response(request)

    def healthz(self, request):
        """
        Returns that the server is alive.
        """
        return HttpResponse("OK")

    def readiness(self, request):
        # Connect to each database and do a generic standard SQL query
        # that doesn't write any data and doesn't depend on any tables
        # being present.
        try:
            from django.db import connections
            for name in connections:
                cursor = connections[name].cursor()
                cursor.execute("SELECT 1;")
                row = cursor.fetchone()
                if row is None:
                    return HttpResponseServerError("db: invalid response")
        except Exception as e:
            logger.exception(e)
            return HttpResponseServerError("db: cannot connect to database.")

        # Call get_stats() to connect to each memcached instance and get it's stats.
        # This can effectively check if each is online.
        try:
            from django.core.cache import caches
            from django.core.cache.backends.memcached import BaseMemcachedCache
            for cache in caches.all():
                if isinstance(cache, BaseMemcachedCache):
                    stats = cache._cache.get_stats()
                    if len(stats) != len(cache._servers):
                        return HttpResponseServerError("cache: cannot connect to cache.")
        except Exception as e:
            logger.exception(e)
            return HttpResponseServerError("cache: cannot connect to cache.")

        return HttpResponse("OK")


class UNICEFSocialAuthExceptionMiddleware(SocialAuthExceptionMiddleware):
    """Middleware to ignore Forgot Password Exceptions"""

    def process_exception(self, request, exception):
        if isinstance(exception, (AuthCanceled, AuthMissingParameter)):
            return HttpResponseRedirect(self.get_redirect_uri(request, exception))
        else:
            raise exception

    def get_redirect_uri(self, request, exception):
        error = request.GET.get('error', None)

        # This is what we should expect:
        # ['AADB2C90118: The user has forgotten their password.\r\n
        # Correlation ID: 7e8c3cf9-2fa7-47c7-8924-a1ea91137ba9\r\n
        # Timestamp: 2018-11-13 11:37:56Z\r\n']
        error_description = request.GET.get('error_description', None)
        if error == "access_denied" and error_description is not None:
            if 'AADB2C90118' in error_description:
                auth_class = UNICEFAzureADB2COAuth2()
                redirect_home = auth_class.get_redirect_uri()
                redirect_url = 'https://login.microsoftonline.com/' + \
                               settings.TENANT_ID + \
                               "/oauth2/v2.0/authorize?p=" + \
                               settings.SOCIAL_PASSWORD_RESET_POLICY + \
                               "&client_id=" + settings.KEY + \
                               "&nonce=defaultNonce&redirect_uri=" + redirect_home + \
                               "&scope=openid+email&response_type=code"
                return redirect_url

        # TODO: In case of password reset the state can't be verified figure out a way to log the user in after reset
        return settings.LOGIN_URL



import logging

from admin_extra_urls.extras import action, ExtraUrlMixin, link
from django import forms
from django.contrib import admin, messages
from django.contrib.admin import ModelAdmin, widgets
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.models import Group
from django.forms import Form
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from .graph import default_group, Synchronizer, SyncResult
from .models import BusinessArea, Region, User
from .sync import load_business_area, load_region

logger = logging.getLogger(__name__)


def admin_reverse(model, page="changelist"):
    return reverse(f"admin:{model._meta.app_label}_{model._meta.model_name}_{page}")


@admin.register(Region)
class RegionAdmin(ExtraUrlMixin, ModelAdmin):
    list_display = ['code', 'name']

    @link()
    def sync(self, request):
        load_region()


@admin.register(BusinessArea)
class BusinessAreaAdmin(ExtraUrlMixin, ModelAdmin):
    list_display = ['code', 'name', 'long_name', 'region', 'country']
    list_filter = ['region', 'country']
    search_fields = ('name',)

    @link()
    def sync(self, request):
        try:
            load_business_area()
        except Exception as e:
            logger.error(e)
            self.message_user(request, str(e), messages.ERROR)


class LoadUsersForm(forms.Form):
    emails = forms.CharField(widget=forms.Textarea)


class FF(Form):
    selection = forms.CharField()


@admin.register(User)
class UserAdmin2(ExtraUrlMixin, UserAdmin):
    list_display = ['username', 'display_name', 'email', 'is_staff',
                    'is_active', 'is_superuser', 'is_linked', 'last_login']
    list_filter = ['is_superuser', 'is_staff', 'is_active']
    search_fields = ['username', 'display_name']
    fieldsets = (
        (None, {'fields': (('username', 'azure_id'), 'password')}),
        (_('Personal info'), {'fields': (('first_name', 'last_name',),
                                         ('email', 'display_name'),
                                         ('job_title',),
                                         )}),
        (_('Permissions'), {'fields': ('is_active', 'is_staff', 'is_superuser',
                                       'groups', 'user_permissions')}),
        (_('Important dates'), {'fields': ('last_login', 'date_joined')}),
    )
    add_fieldsets = (
        (None, {
            'classes': ('wide',),
            'fields': ('username', 'password1', 'password2'),
        }),
    )
    readonly_fields = ('azure_id', 'job_title', 'display_name')

    def is_linked(self, obj):
        return bool(obj.azure_id)

    is_linked.boolean = True

    @action(label='Sync')
    def sync_user(self, request, pk):
        obj = self.get_object(request, pk)
        try:
            syncronizer = Synchronizer()
            syncronizer.sync_user(obj)
        except Exception as e:
            self.message_user(request, str(e), messages.ERROR)

        self.message_user(request, "User synchronized")

    @action(label='Link user')
    def link_user_data(self, request, pk):
        opts = self.model._meta
        ctx = {
            'opts': opts,
            'app_label': 'security',
            'change': True,
            'is_popup': False,
            'save_as': False,
            'has_delete_permission': False,
            'has_add_permission': False,
            'has_change_permission': True,
        }
        obj = self.get_object(request, pk)
        syncronizer = Synchronizer()
        try:
            if request.method == 'POST':
                if request.POST.get('selection'):
                    data = syncronizer.get_user(request.POST.get('selection'))
                    syncronizer.sync_user(obj, data['id'])
                    self.message_user(request, "User linked")
                    return None
                else:
                    ctx['message'] = 'Select one entry to link'

            data = syncronizer.search_users(obj)
            ctx['data'] = data
            return TemplateResponse(request, 'admin/link_user.html', ctx)

        except Exception as e:
            self.message_user(request, str(e), messages.ERROR)

    @link()
    def load(self, request):
        opts = self.model._meta
        ctx = {
            'opts': opts,
            'app_label': 'security',
            'change': True,
            'is_popup': False,
            'save_as': False,
            'has_delete_permission': False,
            'has_add_permission': False,
            'has_change_permission': True,
        }
        if request.method == 'POST':
            form = LoadUsersForm(request.POST)
            if form.is_valid():
                synchronizer = Synchronizer()
                emails = form.cleaned_data['emails'].split()
                total_results = SyncResult()
                for email in emails:
                    result = synchronizer.fetch_users("startswith(mail,'%s')" % email,
                                                      callback=default_group)
                    total_results += result
                self.message_user(request,
                                  f"{len(total_results.created)} users have been created,"
                                  f"{len(total_results.updated)} updated."
                                  f"{len(total_results.skipped)} invalid entries found.")
        else:
            form = LoadUsersForm()
        ctx['form'] = form
        return TemplateResponse(request, 'admin/load_users.html', ctx)


class RoleForm(forms.Form):
    # overwrite_existing = forms.BooleanField(help_text="Overwrite existing entries", required=False)
    business_areas = forms.ModelMultipleChoiceField(queryset=BusinessArea.objects.all(),
                                                    widget=widgets.FilteredSelectMultiple('Services', False)
                                                    )
    user = forms.ModelChoiceField(queryset=User.objects.all())
    group = forms.ModelChoiceField(queryset=Group.objects.all())

# @admin.register(Role)
# class RoleAdmin(ExtraUrlMixin, ModelAdmin):
#     list_display = ['user', 'group', 'business_area']
#     search_fields = ('user',)
#     list_filter = ('group', ('business_area', RelatedFieldComboFilter))
#
#     def has_add_permission(self, request):
#         return False
#
#     @link()
#     def add_grants(self, request):
#         opts = self.model._meta
#         ctx = {
#             'opts': opts,
#             'add': False,
#             'has_view_permission': True,
#             'has_editable_inline_admin_formsets': True,
#             'app_label': opts.app_label,
#             'change': True,
#             'is_popup': False,
#             'save_as': False,
#             'media': self.media,
#             'has_delete_permission': False,
#             'has_add_permission': False,
#             'has_change_permission': True,
#         }
#         if request.method == 'POST':
#             form = RoleForm(request.POST)
#             if form.is_valid():
#                 user = form.cleaned_data.pop('user')
#                 business_areas = form.cleaned_data.pop('business_areas')
#                 group = form.cleaned_data.pop('group')
#                 # overwrite_existing = form.cleaned_data.pop('overwrite_existing')
#
#                 for business_area in business_areas:
#                     Role.objects.update_or_create(user=user,
#                                                   business_area=business_area,
#                                                   group=group)
#                 self.message_user(request, 'ACLs created')
#                 return HttpResponseRedirect(admin_reverse(Role))
#         else:
#             form = RoleForm(initial={})
#         ctx['adminform'] = AdminForm(form,
#                                      [(None, {'fields': ['user',
#                                                          'group',
#                                                          'business_areas']})],
#                                      {})
#         ctx['media'] = self.media + form.media
#         return TemplateResponse(request, 'admin/unicef_security/add_grants.html', ctx)



import base64
import json
import os

from django.conf import settings
from django.urls import reverse
from jwt import decode as jwt_decode, DecodeError, ExpiredSignature
from social_core.backends.azuread_b2c import AzureADB2COAuth2
from social_core.backends.azuread_tenant import AzureADTenantOAuth2
from social_core.exceptions import AuthTokenError


class UNICEFAzureADTenantOAuth2Ext(AzureADTenantOAuth2):
    def user_data(self, access_token, *args, **kwargs):
        response = kwargs.get('response')
        id_token = response.get('id_token')

        # decode the JWT header as JSON dict
        jwt_header = json.loads(
            base64.b64decode(id_token.split('.', 1)[0]).decode()
        )

        # get key id and algorithm
        key_id = jwt_header['kid']
        algorithm = jwt_header['alg']
        verify = os.environ.get('OAUTH2_VERIFY', False)
        key = ''
        try:
            # retrieve certificate for key_id
            if verify:
                certificate = self.get_certificate(key_id)
                key = certificate.public_key()

            return jwt_decode(
                id_token,
                verify=verify,
                key=key,
                algorithms=algorithm,
                audience=self.setting('KEY')
            )
        except (DecodeError, ExpiredSignature) as error:
            raise AuthTokenError(self, error)


class UNICEFAzureADB2COAuth2(AzureADB2COAuth2):
    """UNICEF Azure ADB2C Custom Backend"""

    name = 'unicef-azuread-b2c-oauth2'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.redirect_uri = settings.HOST + reverse('social:complete', kwargs={'backend': 'unicef-azuread-b2c-oauth2'})



import logging

import requests
from django_countries import countries

from . import config
from .graph import SyncResult
from .models import BusinessArea, Region

logger = logging.getLogger(__name__)

c = dict(countries)


def load_region():
    url = "{}businessareas".format(config.INSIGHT_URL)
    response = requests.get(url, headers={'Ocp-Apim-Subscription-Key': config.INSIGHT_SUB_KEY})
    if response.status_code in [401, 403]:
        raise PermissionError('%s - %s: Invalid credentials' % (url, response.status_code))
    data = response.json()['ROWSET']['ROW']
    results = SyncResult()
    regions = set((e['REGION_CODE'], e['REGION_NAME']) for e in data)

    for entry in regions:
        region, created = Region.objects.update_or_create(code=entry[0],
                                                          defaults={
                                                              'name': entry[1]}
                                                          )
        results.log(region, created)
    logger.info(f"Region sync completed: {results}")
    return results


def load_business_area():
    url = "{}businessareas".format(config.INSIGHT_URL)
    response = requests.get(url, headers={'Ocp-Apim-Subscription-Key': config.INSIGHT_SUB_KEY})
    if response.status_code in [401, 403]:
        raise PermissionError('%s - %s: Invalid credentials' % (url, response.status_code))
    data = response.json()['ROWSET']['ROW']
    results = SyncResult()
    for entry in data:
        defaults = {'name': entry['BUSINESS_AREA_NAME'],
                    'country': countries.by_name(entry['BUSINESS_AREA_NAME']),
                    'region': Region.objects.get_or_create(code=entry['REGION_CODE'],
                                                           defaults={
                                                               'name': entry['REGION_NAME']}
                                                           )[0]
                    }
        area, created = BusinessArea.objects.update_or_create(code=entry['BUSINESS_AREA_CODE'],
                                                              defaults=defaults)
        results.log(area, created)
    logger.info(f"BusinessArea sync completed: {results}")
    return results



from django.contrib.auth.models import AbstractUser
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import ugettext as _
from django_countries.fields import CountryField

app_label = 'unicef_security'


class TimeStampedModel:
    last_modify_date = models.DateTimeField(editable=False, blank=True, auto_now_add=True,
                                            auto_now=True)


class Region(models.Model, TimeStampedModel):
    code = models.CharField(_('code'), max_length=10, unique=True)
    name = models.CharField(_('name'), max_length=50, unique=True)

    class Meta:
        app_label = 'unicef_security'

    def __str__(self):
        return f"{self.name}"


class AbstractBusinessArea(models.Model, TimeStampedModel):
    code = models.CharField(_('code'), max_length=10, unique=True)
    name = models.CharField(_('name'), max_length=50, unique=True)
    long_name = models.CharField(_('long name'), max_length=150)
    region = models.ForeignKey(Region, on_delete=models.CASCADE)
    country = CountryField()

    class Meta:
        app_label = 'unicef_security'
        abstract = True
        verbose_name = _('Business Area')
        verbose_name_plural = _('Business Areas')

    def __str__(self):
        return f"{self.name}"


class BusinessArea(AbstractBusinessArea, TimeStampedModel):
    class Meta:
        app_label = 'unicef_security'
        swappable = 'BUSINESSAREA_MODEL'


class User(AbstractUser, TimeStampedModel):
    # business_area = models.ForeignKey(settings.BUSINESSAREA_MODEL,
    #                                   null=True, blank=True,
    #                                   on_delete=models.CASCADE)
    azure_id = models.UUIDField(blank=True, unique=True, null=True)
    job_title = models.CharField(max_length=100, null=True, blank=True)
    display_name = models.CharField(max_length=100, null=True, blank=True)

    class Meta:
        app_label = 'unicef_security'

    @cached_property
    def label(self):
        if self.display_name:
            return self.display_name
        elif self.first_name and self.last_name:
            return f"{self.first_name} {self.last_name}:"
        elif self.first_name:
            return self.first_name
        else:
            return self.username

    def save(self, *args, **kwargs):
        if not self.display_name:
            self.display_name = self.label
        super().save(*args, **kwargs)
